"use strict";
/*
Language: C++
Author: Ivan Sagalaev <maniac@softwaremaniacs.org>
Contributors: Evgeny Stepanischev <imbolk@gmail.com>, Zaven Muradyan <megalivoithos@gmail.com>, Roel Deckers <admin@codingcat.nl>, Sam Wu <samsam2310@gmail.com>, Jordi Petit <jordi.petit@gmail.com>, Pieter Vantorre <pietervantorre@gmail.com>
Category: common, system
*/
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var common_1 = require("../common");
var CPP_PRIMITIVE_TYPES = {
    className: 'keyword',
    begin: '\\b[a-z\\d_]*_t\\b'
};
exports.STRINGS = {
    className: 'string',
    variants: [
        {
            begin: '(u8?|U)?L?"', end: '"',
            illegal: '\\n',
            contains: [common_1.BACKSLASH_ESCAPE]
        },
        {
            begin: '(u8?|U)?R"', end: '"',
            contains: [common_1.BACKSLASH_ESCAPE]
        },
        {
            begin: '\'\\\\?.', end: '\'',
            illegal: '.'
        }
    ]
};
var NUMBERS = {
    className: 'number',
    variants: [
        { begin: '\\b(0b[01\']+)' },
        { begin: '(-?)\\b([\\d\']+(\\.[\\d\']*)?|\\.[\\d\']+)(u|U|l|L|ul|UL|f|F|b|B)' },
        { begin: '(-?)(\\b0[xX][a-fA-F0-9\']+|(\\b[\\d\']+(\\.[\\d\']*)?|\\.[\\d\']+)([eE][-+]?[\\d\']+)?)' }
    ],
    relevance: 0
};
exports.PREPROCESSOR = {
    className: 'meta',
    begin: /#\s*[a-z]+\b/, end: /$/,
    keywords: {
        'meta-keyword': 'if else elif endif define undef warning error line ' +
            'pragma ifdef ifndef include'
    },
    contains: [
        {
            begin: /\\\n/, relevance: 0
        },
        tslib_1.__assign({}, exports.STRINGS, { className: 'meta-string' }),
        {
            className: 'meta-string',
            begin: /<[^\n>]*>/, end: /$/,
            illegal: '\\n',
        },
        common_1.C_LINE_COMMENT_MODE,
        common_1.C_BLOCK_COMMENT_MODE
    ]
};
var FUNCTION_TITLE = common_1.IDENT_RE + '\\s*\\(';
exports.CPP_KEYWORDS = {
    keyword: 'int float while private char catch import module export virtual operator sizeof ' +
        'dynamic_cast|10 typedef const_cast|10 const for static_cast|10 union namespace ' +
        'unsigned long volatile static protected bool template mutable if public friend ' +
        'do goto auto void enum else break extern using asm case typeid ' +
        'short reinterpret_cast|10 default double register explicit signed typename try this ' +
        'switch continue inline delete alignof constexpr decltype ' +
        'noexcept static_assert thread_local restrict _Bool complex _Complex _Imaginary ' +
        'atomic_bool atomic_char atomic_schar ' +
        'atomic_uchar atomic_short atomic_ushort atomic_int atomic_uint atomic_long atomic_ulong atomic_llong ' +
        'atomic_ullong new throw return ' +
        'and or not',
    built_in: 'std string cin cout cerr clog stdin stdout stderr stringstream istringstream ostringstream ' +
        'auto_ptr deque list queue stack vector map set bitset multiset multimap unordered_set ' +
        'unordered_map unordered_multiset unordered_multimap array shared_ptr abort abs acos ' +
        'asin atan2 atan calloc ceil cosh cos exit exp fabs floor fmod fprintf fputs free frexp ' +
        'fscanf isalnum isalpha iscntrl isdigit isgraph islower isprint ispunct isspace isupper ' +
        'isxdigit tolower toupper labs ldexp log10 log malloc realloc memchr memcmp memcpy memset modf pow ' +
        'printf putchar puts scanf sinh sin snprintf sprintf sqrt sscanf strcat strchr strcmp ' +
        'strcpy strcspn strlen strncat strncmp strncpy strpbrk strrchr strspn strstr tanh tan ' +
        'vfprintf vprintf vsprintf endl initializer_list unique_ptr',
    literal: 'true false nullptr NULL'
};
var EXPRESSION_CONTAINS = [
    CPP_PRIMITIVE_TYPES,
    common_1.C_LINE_COMMENT_MODE,
    common_1.C_BLOCK_COMMENT_MODE,
    NUMBERS,
    exports.STRINGS
];
exports.CPlusPlus = {
    name: 'cplusplus',
    aliases: ['c', 'cc', 'h', 'c++', 'h++', 'cpp', 'hpp'],
    keywords: exports.CPP_KEYWORDS,
    illegal: '</',
    contains: EXPRESSION_CONTAINS.concat([
        exports.PREPROCESSOR,
        {
            begin: '\\b(deque|list|queue|stack|vector|map|set|bitset|multiset|multimap|unordered_map|unordered_set|unordered_multiset|unordered_multimap|array)\\s*<', end: '>',
            keywords: exports.CPP_KEYWORDS,
            contains: ['self', CPP_PRIMITIVE_TYPES]
        },
        {
            begin: common_1.IDENT_RE + '::',
            keywords: exports.CPP_KEYWORDS
        },
        {
            // This mode covers expression context where we can't expect a function
            // definition and shouldn't highlight anything that looks like one:
            // `return some()`, `else if()`, `(x*sum(1, 2))`
            variants: [
                { begin: /=/, end: /;/ },
                { begin: /\(/, end: /\)/ },
                { beginKeywords: 'new throw return else', end: /;/ }
            ],
            keywords: exports.CPP_KEYWORDS,
            contains: EXPRESSION_CONTAINS.concat([
                {
                    begin: /\(/, end: /\)/,
                    keywords: exports.CPP_KEYWORDS,
                    contains: EXPRESSION_CONTAINS.concat(['self']),
                    relevance: 0
                }
            ]),
            relevance: 0
        },
        {
            className: 'function',
            begin: '(' + common_1.IDENT_RE + '[\\*&\\s]+)+' + FUNCTION_TITLE,
            returnBegin: true, end: /[{;=]/,
            excludeEnd: true,
            keywords: exports.CPP_KEYWORDS,
            illegal: /[^\w\s\*&]/,
            contains: [
                {
                    begin: FUNCTION_TITLE, returnBegin: true,
                    contains: [common_1.TITLE_MODE],
                    relevance: 0
                },
                {
                    className: 'params',
                    begin: /\(/, end: /\)/,
                    keywords: exports.CPP_KEYWORDS,
                    relevance: 0,
                    contains: [
                        common_1.C_LINE_COMMENT_MODE,
                        common_1.C_BLOCK_COMMENT_MODE,
                        exports.STRINGS,
                        NUMBERS,
                        CPP_PRIMITIVE_TYPES
                    ]
                },
                common_1.C_LINE_COMMENT_MODE,
                common_1.C_BLOCK_COMMENT_MODE,
                exports.PREPROCESSOR
            ]
        },
        {
            className: 'class',
            beginKeywords: 'class struct', end: /[{;:]/,
            contains: [
                { begin: /</, end: />/, contains: ['self'] },
                common_1.TITLE_MODE
            ]
        }
    ])
};
//# sourceMappingURL=cplusplus.js.map